<?php
// $Id: pifr_client_review.inc,v 1.26 2009/07/17 20:46:14 boombatower Exp $
/**
 * @file
 * Provide base review class.
 *
 * Copyright 2008-2009 by Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

class PIFRClientReview {

  protected $CHECKOUT_DIRECTORY;

  protected $REVIEW_DIRECTORY;

  protected $database;

  protected $vcs;

  protected $test;

  protected $filePath;

  protected $adminPassword;

  protected $error = FALSE;

  public function __construct(array $test) {
    $this->CHECKOUT_DIRECTORY = file_directory_path() . '/checkout';
    $this->REVIEW_DIRECTORY = file_directory_path() . '/review';
    $this->test = $test;

    // Require database and version control system interace.
    require_once drupal_get_path('module', 'pifr_client') . '/review/db/db.inc';
    require_once drupal_get_path('module', 'pifr_client') . '/review/vcs/vcs.inc';

    // Create database interface implementation.
    require_once drupal_get_path('module', 'pifr_client') . '/review/db/' . PIFR_CLIENT_DB . '.inc';
    $database = 'PIFRClientReviewDB_' . PIFR_CLIENT_DB;
    $this->database = new $database;

    // Create version control system interface implementations.
    $vcs_types = $this->getRepositoryTypes();
    foreach ($vcs_types as $vcs_type) {
      require_once drupal_get_path('module', 'pifr_client') . "/review/vcs/$vcs_type.inc";
      $vcs = 'PIFRClientReviewVCS_' . $vcs_type;
      $this->vcs[$vcs_type] = new $vcs;
    }
  }

  protected function getRepositoryTypes($type = NULL) {
    $types = array();
    $types['core'] = $this->test['core']['repository']['type'];

    if (isset($this->test['module'])) {
      $types['module'] = $this->test['module']['repository']['type'];
    }

    if ($type) {
      return $types[$type];
    }

    return array_unique($types);
  }

  public function hasError() {
    return $this->error !== FALSE;
  }

  public function getError() {
    return $this->error;
  }

  protected function setError($error = array()) {
    if (is_int($error)) {
      $this->error['code'] = $error;
    }
    else {
      $this->error = $error;
    }
  }

  public function run() {
    // Init statics.
    watchdog('pifr_client', 'review.initStatics()');
    $this->setWorkingDirectory(TRUE);
    $this->setDBPrefix(TRUE);
    $this->setPath(TRUE);

    $operations = array('setup', 'fetchFile', 'checkout', 'checkPatch', 'applyPatch',
                         'checkSyntax', 'checkFormat', 'install', 'runTests');

    $code = 1;
    foreach ($operations as $operation) {
      watchdog('pifr_client', 'review.' . $operation . '()');
      $this->$operation();

      if ($this->hasError()) {
        $error = $this->getError();
        if (!is_int($error) && !is_string($error)) {
          $error = var_export($error, TRUE);
        }
        watchdog('pifr_client', 'review.setError("' . $operation . '", ' . $code . ', "' . $error . '")');
        $this->setError($code);
        break;
      }

      $code++;
    }

    // In case any operation failed reset data members.
    watchdog('pifr_client', 'review.resetStatics()');
    $this->setWorkingDirectory(TRUE);
    $this->setDBPrefix(TRUE);
    $this->setPath(TRUE);
  }

  protected function setup() {
    if (file_exists($this->CHECKOUT_DIRECTORY)) {
      if (file_exists($this->CHECKOUT_DIRECTORY . '/sites/default')) {
        // Change permissions on settings.php which is read-only after installation.
        chmod($this->CHECKOUT_DIRECTORY . '/sites/default', 0777);
      }

      // Remove checkout directory.
      if (!$this->exec('rm -r ' . $this->CHECKOUT_DIRECTORY)) {
        $this->setError(array('@reason' => t('failed to clear checkout directory')));
        return;
      }
    }

    // Create checkout directory.
    if (!file_check_directory($this->CHECKOUT_DIRECTORY, FILE_CREATE_DIRECTORY)) {
      $this->setError(array('@reason' => t('failed to create checkout directory')));
      return;
    }

    // Create review directory.
    if (!file_check_directory($this->REVIEW_DIRECTORY, FILE_CREATE_DIRECTORY)) {
      $this->setError(array('@reason' => t('failed to create review directory')));
      return;
    }

    // Ensure we can run site by using symbolic link.
    if (!file_exists('./checkout')) {
      $this->setError(array('@reason' => t('failed to find symbolic link needed to invoke checkout')));
      return;
    }

    // Ensure that checkout database exists.
    global $db_url, $db_prefix;

    // Escape db prefix.
    $db_prefix = '';

    if (!is_array($db_url) || !array_key_exists('pifr_checkout', $db_url)) {
      $this->setError(array('@reason' => t('checkout database not found')));
      return;
    }

    // Clear database.
    if (!$this->database->dropDatabase()) {
      $this->setError(array('@reason' => t('failed to drop checkout database')));
      return;
    }

    // Create database.
    if (!$this->database->createDatabase()) {
      $this->setError(array('@reason' => t('failed to create checkout database')));
      return;
    }
  }

  /**
   * Fetch the file for review, if specified.
   */
  protected function fetchFile() {
    if (!isset($this->test['file_url'])) {
      return; // No file specified, ignore this step.
    }

    // Request file from projet server.
    $response = drupal_http_request($this->test['file_url']);
    if ($response->code != 200) {
      $this->setError(array('@reason' => t('failed to retrieve file from project client')));
      return;
    }

    // Save the file data.
    $this->filePath = $this->REVIEW_DIRECTORY . '/' . basename($this->test['file_url']);
    if (!file_save_data($response->data, $this->filePath, FILE_EXISTS_REPLACE)) {
      $this->setError(array('@reason' => t('failed to save file data')));
      return;
    }
  }

  protected function checkout() {
    // Checkout Drupal core.
    $type = $this->getRepositoryTypes('core');
    if (!$this->vcs[$type]->checkout(basename($this->CHECKOUT_DIRECTORY), $this->test['core']['repository']['url'],
                                      $this->test['core']['branch_identifier'])) {
      $this->setError(array('@reason' => t('failed to checkout Drupal core')));
    }
    $this->setWorkingDirectory();

    if ($this->hasError()) {
      return; // Need to revert working directory before continuing.
    }

    // Checkout module if specified.
    if (isset($this->test['module'])) {
      $directory = $this->CHECKOUT_DIRECTORY . '/sites/default/modules';
      if (!file_check_directory($directory, FILE_CREATE_DIRECTORY)) {
        $this->setError(array('@reason' => t('failed to create modules directory')));
        return;
      }
      chdir($this->CHECKOUT_DIRECTORY . '/sites/default/modules');

      $type = $this->$this->getRepositoryTypes('module');
      $directory = basename($this->test['module']['repository']['url']);
      if (!$this->vcs[$type]->checkout($directory, $this->test['module']['repository']['url'],
                                                  $this->test['module']['branch_identifier'])) {
        $this->setError(array('@reason' => t('failed to checkout module')));
      }
      $this->setWorkingDirectory();

      if ($this->hasError()) {
        return; // Need to revert working directory before continuing.
      }
    }
  }

  /**
   * Check patch for non-unix line endings.
   */
  protected function checkPatch() {
    if (!isset($this->test['file_url'])) {
      return; // No file specified, ignore this step.
    }

    $contents = file_get_contents($this->filePath);
    if (strpos($contents, "\r")) {
      $this->setError();
    }
  }

  protected function applyPatch() {
    if (!isset($this->test['file_url'])) {
      return; // No file specified, ignore this step.
    }

    $type = $this->getRepositoryTypes(isset($this->test['module']) ? 'module' : 'core');
    $file = realpath($this->filePath);
    $directory = $this->getBaseDirectory();

    chdir($directory);
    if (!$this->vcs[$type]->apply($file)) {
      $this->setError();
    }
    $this->setWorkingDirectory();
  }

  protected function checkSyntax() {
    if (!isset($this->test['file_url'])) {
      return; // No file specified, ignore this step.
    }

    // Get a list of all the files to which code has been added or removed.
    $type = $this->getRepositoryTypes(isset($this->test['module']) ? 'module' : 'core');
    $files = $this->vcs[$type]->getChangedFiles($this->filePath);

    $directory = $this->getBaseDirectory();
    chdir($directory);
    foreach ($files as $file) {
      if (file_exists($file) && !$this->exec('php -l -f ./' . $file)) {
        $this->setError(array('@filename' => basename($file)));
      }
    }
    $this->setWorkingDirectory();
  }

  protected function checkFormat() {
    // Do nothing.
  }

  protected function install() {
    // Make a copy of the settings file to ensure there is no permission issue.
    $default_settings = $this->CHECKOUT_DIRECTORY . '/sites/default/default.settings.php';
    $settings = $this->CHECKOUT_DIRECTORY . '/sites/default/settings.php';
    copy($default_settings, $settings);

    require_once drupal_get_path('module', 'pifr_client') . '/review/browser.inc';
    $this->setPath();

    // Install Drupal.
    $b = new PIFRBrowser();

    // Step: Select an installation profile.
    // Step: Choose language.
    $b->drupalGet('install.php', array('query' => 'profile=default&locale=en'));

    // Step: Database configuration.
    $db_info = $this->getDatabaseInformation();
    $edit = array();
    if ($b->xpath('//input[@name="driver"]')) {
      $edit['driver'] = PIFR_CLIENT_DB;
    }
    $edit['database'] = $db_info['name'];
    $edit['username'] = $db_info['username'];
    $edit['password'] = $db_info['password'];
    $edit['host'] = $db_info['host'];
    $edit['db_prefix'] = '';
    $b->drupalPost(NULL, $edit, t('Save and continue'));

    // Step: Site configuration.
    $edit = array();
    $edit['site_name'] = 'checkout';
    $edit['site_mail'] = 'admin@example.com';
    $edit['account[name]'] = 'admin';
    $edit['account[mail]'] = 'admin@example.com';
    $edit['account[pass][pass1]'] = $this->adminPassword = $b->randomName(12);
    $edit['account[pass][pass2]'] = $this->adminPassword;
    $edit['update_status_module[1]'] = FALSE;
    $b->drupalPost(NULL, $edit, t('Save and continue'));

    // Step: Finished.
    $b->assertText(t('Drupal installation complete'));

    $this->setPath(TRUE);

    // Make sure that site installed correctly.
    if (($b->results['#exception'] + $b->results['#fail']) != 0) {
      $this->setError();
    }
  }

  protected function getDatabaseInformation() {
    global $db_url;

    $url = parse_url($db_url['pifr_checkout']);
    $db = array();
    $db['name'] = substr(urldecode($url['path']), 1);
    $db['username'] = urldecode($url['user']);
    $db['password'] = urldecode($url['pass']);
    $db['host'] = urldecode($url['host']);
    return $db;
  }

  protected function runTests() {
    global $pifr_xml_results;

    if (!$this->database->setVariable('simpletest_clear_results', FALSE)) {
      $this->setError(array('@reason' => t('failed to set SimpleTest variable in checkout database')));
    }

    // Login to Drupal HEAD and enable the SimpleTest module.
    $this->setPath();

    $b = new PIFRBrowser();

    // Login as admin.
    $user = new stdClass();
    $user->name = 'admin';
    $user->pass_raw = $this->adminPassword;
    $b->drupalLogin($user);

    // Enable SimpleTest.
    $edit = array();
    $edit['modules[Core][simpletest][enable]'] = TRUE;
    $b->drupalPost('admin/build/modules', $edit, t('Save configuration'));

    $this->setPath(TRUE);

    if (($b->results['#exception'] + $b->results['#fail']) != 0) {
      $this->setError(array('@reason' => t('failed to enable SimpleTest')));
      return;
    }

    // Run tests from command line.
    chdir($this->CHECKOUT_DIRECTORY);

    $url = escapeshellarg(url('', array('absolute' => TRUE)) . 'checkout');
    $tests = (PIFR_DEBUG ? '--class BlockTestCase' : '--all');
    if (!$this->exec(PIFR_CLIENT_PHP . ' ./scripts/run-tests.sh --concurrency ' . PIFR_CLIENT_CONCURRENCY .
                     ' --php ' . PIFR_CLIENT_PHP . ' --url ' . $url . ' ' . $tests)) {
      $this->setError(array('@reason' => t('failed to run tests')));
    }

    $this->setWorkingDirectory();
  }

  public function getResults() {
    $details = ($this->error ? $this->error : array());
    unset($details['code']);

    $result = array(
      'test_id' => $this->test['test_id'],
      'code' => ($this->error ? $this->error['code'] : -1),
      'details' => $details,
      'pass' => 0,
      'fail' => 0,
      'exception' => 0,
      'result_details' => array(),
    );

    if ($result['code'] != -1) {
      return $result;
    }

    // Cycle through assertions to fill in results.
    $assertions = $this->database->getAssertions();
    while ($assertion = db_fetch_array($assertions)) {
      // Add up summary totals for entire test suite.
      $result[$assertion['status']]++;

      if (!isset($result['result_details'][$assertion['test_class']])) {
        $result['result_details'][$assertion['test_class']] = array(
          'test_name' => $assertion['test_class'], // TODO Get test name.
          'pass' => 0,
          'fail' => 0,
          'exception' => 0,
          'assertions' => array(),
        );
      }

      // Add up summary totals for test class.
      $result['result_details'][$assertion['test_class']][$assertion['status']]++;

      if ($assertion['status'] != 'pass' && count($result['result_details']) < PIFR_CLIENT_MAX_ASSERTIONS) {
        // Store non-pass assertion until reaching maximum.
        unset($assertion['message_id']);
        unset($assertion['test_id']);
        $assertion['file'] = basename($assertion['file']);
        $result['result_details'][$assertion['test_class']]['assertions'][] = $assertion;

        watchdog('pifr_client', 'Assertion: ' . var_export($assertion, TRUE));
      }
    }

    // Remove class name keys.
    $result['result_details'] = array_values($result['result_details']);

    // The next two codes beyond the run() loop.
    $result['code'] = ($result['fail'] + $result['exception'] ? 10 : 11);
    return $result;
  }

  public static function exec($command) {
    $output = array();
    exec($command, $output, $status);
    if ($status != 0) {
      watchdog('pifr_client', 'Command: "@command"; Result: "@result".',
               array('@command' => $command, '@result' => implode("\n", $output)));
      return FALSE;
    }
    return TRUE;
  }

  protected function setWorkingDirectory($reset = FALSE) {
    static $original_directory;

    if ($reset) {
      $original_directory = getcwd();
    }

    chdir($original_directory);
  }

  protected function setDBPrefix($reset = FALSE) {
    global $db_prefix;
    static $original_db_prefix;

    if ($reset) {
      $original_db_prefix = $db_prefix;
    }

    $db_prefix = $original_db_prefix;
  }

  protected function getBaseDirectory() {
    $directory = $this->CHECKOUT_DIRECTORY;
    if (isset($this->test['module'])) {
      $directory .= '/sites/default/modules/' . basename($this->test['module']['repository']['url']);
    }
    return $directory;
  }

  /**
   * Set the original, Drupal set, or base_url and base_path to /checkout.
   *
   * @param boolean $original Original path.
   */
  protected function setPath($reset = FALSE) {
    global $base_url, $base_path;
    static $base_url_original, $base_path_original;

    if (!isset($base_url_original) || !isset($base_path_original)) {
      $base_url_original = $base_url;
      $base_path_original = $base_path;
    }

    if ($reset) {
      $base_url = $base_url_original;
      $base_path = $base_path_original;
    }
    else {
      $base_url .= '/checkout';
      $base_path = '/checkout';
    }
  }
}
