<?php
// $Id: pifr_server.result.inc,v 1.14 2009/04/17 05:56:49 boombatower Exp $
/**
 * @file
 * Provide test result functions.
 *
 * Copyright 2008-2009 by Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/**
 * Get the result tree for a particular result.
 *
 * @param integer $result_id Result ID.
 * @return array Result tree.
 *   Summary (pifr_result)
 *     Test summary (pifr_result_detail)
 *       Non-pass assertions (pifr_result_detail_assertion)
 */
function pifr_server_result_get($result_id) {
  $result = db_fetch_array(db_query('SELECT *
                                     FROM {pifr_result}
                                     WHERE result_id = %d', $result_id));
  $result['details'] = unserialize($result['details']);

  $result_details = db_query('SELECT *
                              FROM {pifr_result_detail}
                              WHERE result_id = %d', $result_id);
  $result['result_details'] = array();
  while ($result_detail = db_fetch_array($result_details)) {
    $result_detail_assertions = db_query('SELECT *
                                          FROM {pifr_result_detail_assertion}
                                          WHERE result_detail_id = %d', $result_detail['result_detail_id']);
    $result_detail['assertions'] = array();
    while ($result_detail_assertion = db_fetch_array($result_detail_assertions)) {
      $result_detail['assertions'][] = $result_detail_assertion;
    }
    $result['result_details'][] = $result_detail;
  }
  return $result;
}

/**
 * Get all the results for a particular test.
 *
 * @param integer $test_id Test ID.
 * @return array Set of results keyed by result type. The results will be in
 *   order of increasing type ID.
 * @see pifr_server_result_get().
 */
function pifr_server_result_get_all($test_id) {
  $results = db_query('SELECT result_id
                       FROM {pifr_result}
                       WHERE test_id = %d
                       ORDER BY type ASC', $test_id);

  $test_results = array();
  while ($result_id = db_result($results)) {
    $result = pifr_server_result_get($result_id);
    $test_results[$result['type']] = $result;
  }
  return $test_results;
}

function pifr_server_result_get_status($test_id) {
  $results = pifr_server_result_get_all($test_id);
  foreach ($results as $result) {
    if ($result['code'] != PIFR_SERVER_TEST_RESULT_PASS) {
      return FALSE;
    }
  }
  return TRUE;
}

function pifr_server_result_get_summary($test_id, $colored = FALSE) {
  $message = pifr_server_result_get_summary_message(PIFR_SERVER_TEST_RESULT_PASS);
  $details = array();

  $results = pifr_server_result_get_all($test_id);
  $pass = TRUE;
  foreach ($results as $result) {
    if ($result['code'] != PIFR_SERVER_TEST_RESULT_PASS) {
      $message = pifr_server_result_get_summary_message($result['code']);

      if ($result['code'] == PIFR_SERVER_TEST_RESULT_FAIL) {
        $details = array(
          '@pass' => $result['pass'],
          '@fail' => $result['fail'],
          '@exception' => $result['exception'],
        );
      }
      else {
        $details = $result['details'];
      }
      $details['@environment'] = pifr_server_client_environment($result['type']);
      $pass = FALSE;
      break;
    }
  }

  if (!$results) {
    $message = 'Pending';
    $details = array();
    $colored = FALSE;
  }
  else if ($pass && count($results) < PIFR_SERVER_CLIENT_ENVIRONMENT_COUNT) {
    $message = 'Passed on @num of @count environments.';
    $details['@num'] = count($results);
    $details['@count'] = PIFR_SERVER_CLIENT_ENVIRONMENT_COUNT;
  }

  if ($colored) {
    $class = ($pass ? 'pifr-ok' : 'pifr-error');
    return '<div id="pifr-summary" class="' . $class . '">' . t($message, $details) . '</div>';
  }
  return t($message, $details);
}

function pifr_server_result_get_summary_message($code) {
  switch ($code) {
    case PIFR_SERVER_TEST_RESULT_SETUP:
      return 'Setup environment: @reason.';
    case PIFR_SERVER_TEST_RESULT_FETCH:
      return 'Fetch test file: @reason.';
    case PIFR_SERVER_TEST_RESULT_CVS:
      return 'Repository checkout: @reason.';
    case PIFR_SERVER_TEST_RESULT_PATCH:
      return 'Invalid patch format.';
    case PIFR_SERVER_TEST_RESULT_APPLY:
      return 'Unable to apply patch.';
    case PIFR_SERVER_TEST_RESULT_SYNTAX:
      return 'Invalid PHP syntax in @filename.';
    case PIFR_SERVER_TEST_RESULT_FORMAT:
      return '-- not implemented --';
    case PIFR_SERVER_TEST_RESULT_INSTALL:
      return 'Failed to install Drupal on @environment.';
    case PIFR_SERVER_TEST_RESULT_TEST:
      return 'Run tests on @environment: @reason.';
    case PIFR_SERVER_TEST_RESULT_FAIL:
      return 'Failed on @environment, with: @pass pass, @fail fail, and @exception exception';
    case PIFR_SERVER_TEST_RESULT_PASS:
      return 'Passed on all environments.';
    default:
      return 'Unknown';
  }
}

function pifr_server_result_get_summary_message_minimal($code) {
  switch ($code) {
    case PIFR_SERVER_TEST_RESULT_SETUP:
      return 'Setup environment.';
    case PIFR_SERVER_TEST_RESULT_FETCH:
      return 'Fetch test file.';
    case PIFR_SERVER_TEST_RESULT_CVS:
      return 'Repository checkout.';
    case PIFR_SERVER_TEST_RESULT_PATCH:
      return 'Invalid patch format.';
    case PIFR_SERVER_TEST_RESULT_APPLY:
      return 'Unable to apply patch.';
    case PIFR_SERVER_TEST_RESULT_SYNTAX:
      return 'Invalid PHP syntax.';
    case PIFR_SERVER_TEST_RESULT_FORMAT:
      return '-- not implemented --';
    case PIFR_SERVER_TEST_RESULT_INSTALL:
      return 'Failed to install Drupal.';
    case PIFR_SERVER_TEST_RESULT_TEST:
      return 'Run tests.';
    case PIFR_SERVER_TEST_RESULT_FAIL:
      return 'Failed on tests.';
    case PIFR_SERVER_TEST_RESULT_PASS:
      return 'Passed on all environments.';
    default:
      return 'Unknown';
  }
}

/**
 * Clear all test result records (includes tree).
 *
 * @param integer $test_id Test ID.
 * @see pifr_server_result_get().
 */
function pifr_server_result_clear($test_id) {
  $results = pifr_server_result_get_all($test_id);
  foreach ($results as $result) {
    foreach ($result['details'] as $detail) {
      db_query('DELETE FROM {pifr_result_detail_assertion}
                WHERE result_detail_id = %d', $detail['result_detail_id']);
    }
    db_query('DELETE FROM {pifr_result_detail}
              WHERE result_id = %d', $result['result_id']);
  }
  db_query('DELETE FROM {pifr_result}
            WHERE test_id = %d', $test_id);
}

/**
 * Save the specified result. The data will be saved in the three result
 * tables: pifr_result, pifr_result_detail, pifr_result_detail_assertion.
 *
 * @param array $result Result to save.
 * @return interger Result ID of saved result.
 */
function pifr_server_result_save(array $result) {
  db_query("INSERT INTO {pifr_result} (test_id, type, code, details, pass, fail, exception)
            VALUES (%d, %d, %d, '%s', %d, %d, %d)",
            $result['test_id'], $result['type'], $result['code'], $result['details'],
            $result['pass'], $result['fail'], $result['exception']);
  $result_id = db_last_insert_id('pifr_result', 'result_id');

  foreach ($result['result_details'] as $result_detail) {
    db_query("INSERT INTO {pifr_result_detail} (result_id, test_name, pass, fail, exception)
            VALUES (%d ,'%s', %d, %d, %d)",
            $result_id, $result_detail['test_name'], $result_detail['pass'],
            $result_detail['fail'], $result_detail['exception']);
    $result_detail_id = db_last_insert_id('pifr_result_detail', 'result_detail_id');

    foreach ($result_detail['assertions'] as $assertion) {
      db_query("INSERT INTO {pifr_result_detail_assertion}
                (result_detail_id, status, message, message_group, function, line, file)
                VALUES (%d, '%s', '%s', '%s', '%s', '%s', '%s')",
                $result_detail_id, $assertion['status'], $assertion['message'], $assertion['message_group'],
                $assertion['function'], $assertion['line'], $assertion['file']);
    }
  }

  return $result_id;
}

/**
 * Check if the result is a pass.
 *
 * @param $result Result object.
 * @return boolean Pass.
 */
function pifr_server_result_is_pass($result) {
  return ($result['fail'] + $result['exception']) == 0;
}
