<?php
// $Id: pifr_server.xmlrpc.inc,v 1.27 2009/04/14 07:51:41 boombatower Exp $
/**
 * @file
 * Provide XML-RPC callbacks and invoke functions.
 *
 * Copyright 2008-2009 by Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/**
 * Queue a the batch of files for testing.
 *
 * @param string $client_key Client key requesting batch to be queued.
 * @param array $batch Array of file and project data used to queue files.
 * <code>
 *   $batch = array(
 *     'branches' => array( [referenced branches (may also be tested)] ),
 *     'files' => array( [files to be tested] ),
 *     'projects' => array( [referenced projects] ),
 *   );
 *
 *   $branch = array(
 *     'branch_nid', // Release NID.
 *     'project_nid', // Project NID.
 *     'core_branch_nid', // (Optional) Compatible core branch release NID.
 *     'identifier', // Repository branch identifier.
 *     'test', // Request branch to be tested.
 *   );
 *
 *   $file = array(
 *     'branch_nid', // Branch release NID.
 *     'file_url', // Complete URL to file.
 *     'issue_nid', // Issue NID.
 *     'issue_cid', // Issue CID.
 *   );
 *
 *   $project = array(
 *     'project_nid', // Project NID.
 *     'name', // Short project name.
 *     'type', // Project type, PIFR_SERVER_PROJECT_TYPE_*.
 *     'repository_type', // Repository type (currently only CVS).
 *     'repository_url', // Repository URL, pserver:anonymous:anonymous@cvs.drupal.org:/cvs/drupal.
 *   );
 * </code>
 * @return array Assigned test IDs, or array with response code, PIFR_RESPONSE_*.
 * <code>
 *   $response = array(
 *     'branches' => array( [Test IDs of branches in order of increasing branch NID] ),
 *     'files' => array( [Test IDs of files in order ascending order by filename] ),
 *   );
 * </code>
 *
 * OR
 *
 * <code>
 *   $test = array(
 *     'response' // Response code: PIFR_RESPONSE_*.
 *   );
 * </code>
 */
function pifr_server_xmlrpc_queue($client_key, $batch) {
  pifr_debug('XML-RPC batch(): received.');

  // Check if valid project client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_PROJECT);

  if ($client) {
    pifr_debug('XML-RPC batch(): valid project client.');

    // Store project data.
    $projects = pifr_server_xmlrpc_project_save($batch['projects'], $client['client_id']);

    // Find and store core branches.
    $core_branches = pifr_server_xmlrpc_branch_save($batch['branches'], $projects);

    // Find the non-core branches.
    $branches = $batch['branches'];
    foreach ($branches as $key => $branch) {
      if (array_key_exists($branch['branch_nid'], $core_branches)) {
        unset($branches[$key]);
      }
    }

    // Store non-core branch data.
    $branches = pifr_server_xmlrpc_branch_save($branches, $projects, $core_branches);

    // Merge core and non-core branches.
    $branches += $core_branches;

    // Store file data.
    $files = pifr_server_xmlrpc_file_save($batch['files'], $branches);

    // Postpone tests if branch is currently failing.
    foreach ($branches as $branch) {
      if (!pifr_server_result_get_status($branch['test_id'])) {
        // Postpone all tests that have been queued.
        pifr_server_file_branch_update($branch['branch_id'], PIFR_SERVER_TEST_STATUS_QUEUED, PIFR_SERVER_TEST_STATUS_POSTPONED);
      }
    }

    // Sort branches by NID and files by file name.
    ksort($branches);
    ksort($files);

    // Send back test IDs.
    $response = array(
      'branches' => array(),
      'files' => array(),
    );
    foreach ($branches as $branch) {
      $response['branches'][] = $branch['test_id'];
    }
    foreach ($files as $file) {
      $response['files'][] = $file['test_id'];
    }

    return $response;
  }
  else {
    pifr_debug('XML-RPC batch(): invalid project client.');
    return array(
      'response' => PIFR_RESPONSE_INVALID_SERVER,
    );
  }
}

/**
 * Save specified projects.
 *
 * @param array $projects Projects to save.
 * @param integer $client_id Client ID that send projects.
 * @return array Saved projects.
 */
function pifr_server_xmlrpc_project_save(array $projects, $client_id) {
  $saved = array();
  foreach ($projects as $key => $project) {
    $project['client_id'] = $client_id;
    $saved[$project['project_nid']] = pifr_server_project_save($project);
  }
  return $saved;
}

/**
 * Save specified branches.
 *
 * @param array $branches Branches to save.
 * @param array $projects Projects that branches relate to.
 * @param mixed $core_branches Core branches related to branches or FALSE to save core branches.
 * @return array Saved branches.
 */
function pifr_server_xmlrpc_branch_save(array $branches, array $projects, $core_branches = FALSE) {
  $saved = array();
  foreach ($branches as $branch) {
    if (!isset($projects[$branch['project_nid']])) {
      pifr_debug(t('XML-RPC batch(): missing project data for branch: (b: @identifier; p: @project_nid).',
                   array('@identifier' => $branch['identifier'], '@project_nid' => $branch['project_nid'])));
      continue;
    }
    if ($core_branches && !isset($core_branches[$branch['core_branch_nid']])) {
      pifr_debug(t('XML-RPC batch(): missing core branch data for branch: (b: @identifier; p: @project_nid).',
                   array('@identifier' => $branch['identifier'], '@project_nid' => $branch['project_nid'])));
      continue;
    }

    if (!$core_branches && isset($branch['core_branch_nid'])) {
      // Only save core branches.
      continue;
    }

    $branch['project_id'] = $projects[$branch['project_nid']]['project_id'];
    $branch['core_branch_id'] = ($core_branches ? $core_branches[$branch['core_branch_nid']]['branch_id'] : 0);
    $saved[$branch['branch_nid']] = pifr_server_branch_save($branch, TRUE);

    // Update test record.
    if ($branch['test']) {
      $test = pifr_server_test_get($saved[$branch['branch_nid']]['test_id']);
      pifr_server_test_queue($test);
    }
  }
  return $saved;
}

/**
 * Save specified files.
 *
 * @param array $files Files to save.
 * @param array $branches Branches that files relate to.
 * @return array Saved files.
 */
function pifr_server_xmlrpc_file_save(array $files, array $branches) {
  $saved = array();
  foreach ($files as $file) {
    if (!isset($branches[$file['branch_nid']])) {
      pifr_debug(t('XML-RPC batch(): missing branch data for file: (u: @file_url; n: @issue_nid).',
                   array('@file_url' => $file['file_url'], '@issue_nid' => $file['issue_nid'])));
      continue;
    }

    $file['branch_id'] = $branches[$file['branch_nid']]['branch_id'];
    $file['file_name'] = basename($file['file_url']);
    $saved[$file['file_name']] = $file = pifr_server_file_save($file);

    // Update test record.
    $test = pifr_server_test_get($file['test_id']);
    pifr_server_test_queue($test);
  }
  return $saved;
}

/**
 * Request the next file for testing. (Called by PIFR client)
 *
 * @param string $client_key Client key requesting next test.
 * @return array Test information, empty array, or array with response code, PIFR_RESPONSE_*.
 * <code>
 *   $test = array(
 *     'test_id', // Unique test ID.
 *     'core' => array(
 *       'repository' => array(
 *          'type', // Textual representation of repository type.
 *          'url', // CVS will only require URL.
 *       ),
 *       'branch_identifier', // Branch identifier.
 *     ),
 *     'module' => array( // (Optional) If included then module will be tested.
 *       'repository' => array(
 *         'type', // Textual representation of repository type.
 *         'url', // CVS will only require URL.
 *       ),
 *       'branch_identifier', // Branch identifier.
 *     ),
 *     'file_url', // (Optional) If included then file will be downloaded and applied to code.
 *   );
 * </code>
 *
 * OR
 *
 * <code>
 *   $test = array(
 *     'response' // Response code: PIFR_RESPONSE_*.
 *   );
 * </code>
 */
function pifr_server_xmlrpc_next($client_key) {
  pifr_debug('XML-RPC next(): received.');

  // Check if valid test client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_TEST);

  if ($client) {
    pifr_debug('XML-RPC next(): valid test client.');

    return pifr_server_test_get_next($client);
  }
  else {
    pifr_debug('XML-RPC next(): invalid test client.');
    return array(
      'response' => PIFR_RESPONSE_INVALID_SERVER,
    );
  }
  return array(
    'response' => PIFR_RESPONSE_ACCEPTED,
  );
}

/**
 * Report test results. (Called by PIFR client)
 *
 * @param string $client_key Client key sending test results.
 * @param array $result Test result.
 * <code>
 *   $result = array(
 *     'test_id', // Unique test ID.
 *     'code', // Summary result code, PIFR_SERVER_TEST_RESULT_*.
 *     'details', // Array of details (parameters for t()).
 *     'pass', // Total number of pass assertions.
 *     'fail', // Total number of fail assertions.
 *     'exception', // Total number of exception assertions.
 *     'result_details' => array( [summary of test and non-pass assertions] ),
 *   );
 *
 *   $result_detail = array(
 *     'test_name', // Name of test.
 *     'pass', // Total number of pass assertions.
 *     'fail', // Total number of fail assertions.
 *     'exception', // Total number of exception assertions.
 *     'assertions' => array( [non-pass assertions] ),
 *   );
 *
 *   $assertion = array(
 *     'status', // Assertion status: 'fail' or 'exception'.
 *     'message', // Descriptive assertion message.
 *     'message_group', // Message group.
 *     'function', // Function in which the assertion was triggered.
 *     'line', // Line at which the assertion was triggered.
 *     'file', // File in which the assertion was triggered.
 *   );
 * </code>
 * @return Response code: PIFR_RESPONSE_*.
 */
function pifr_server_xmlrpc_result($client_key, array $result) {
  pifr_debug('XML-RPC result(): received.');

  // Check if valid test client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_TEST);

  if ($client) {
    pifr_debug('XML-RPC result(): valid test client.');

    // Ensure that test was assigned to client.
    $test = pifr_server_test_get($result['test_id']);
    if ($test['status'] != PIFR_SERVER_TEST_STATUS_SENT || $test['client_id'] != $client['client_id']) {
      watchdog('pifr_server', 'Test not assigned to client: (t: @test_id, c: @client_id)',
               array('@test_id' => $test['test_id'], '@client_id' => $client['client_id']), WATCHDOG_ALERT);
      return PIFR_RESPONSE_DENIED;
    }

    pifr_server_test_result($test, $result, $client);
  }
  else {
    pifr_debug('XML-RPC result(): invalid test client.');
    return PIFR_RESPONSE_INVALID_SERVER;
  }
  return PIFR_RESPONSE_ACCEPTED;
}

/**
 * Retrieve test results. (Called by PIFT)
 *
 * @param string $client_key Client key requesting batch to be queued.
 * @param integer $since GMT Unix timestamp to query test results after.
 * @return array Test results, empty array, or array with response code, PIFR_RESPONSE_*.
 * <code>
 *   $batch = array(
 *     'results' => array( [results of tests] ),
 *     'next', // More results are available, query using the provided GMT Unix timestamp.
 *   );
 *
 *   $result = array(
 *     'test_id', // Unique ID representing the test.
 *     'pass', // TRUE or FALSE.
 *     'message', // Message describing results of test.
 *   );
 * </code>
 *
 * OR
 *
 * <code>
 *   $test = array(
 *     'response' // Response code: PIFR_RESPONSE_*.
 *   );
 * </code>
 */
function pifr_server_xmlrpc_retrieve($client_key, $since) {
  pifr_debug('XML-RPC retrieve(): received.');

  if (!PIFR_SERVER_REPORT) {
    pifr_debug('XML-RPC retrieve(): ignored (reporting disabled).');
    return array(
      'response' => PIFR_RESPONSE_ACCEPTED,
    );
  }

  // Check if valid test client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_PROJECT);

  if ($client) {
    pifr_debug('XML-RPC retrieve(): valid project client.');

    $batch = array(
      'results' => array(),
      'next' => FALSE,
    );

    $count = 0;
    $since = pifr_server_xmlrpc_time_local($since);
    $tests = pifr_server_test_get_since($since);
    foreach ($tests as $test) {
      if ($count >= PIFR_SERVER_BATCH_MAX_COUNT) {
        $batch['next'] = pifr_server_xmlrpc_time_gmt($test['last_tested']);
        break;
      }

      $batch['results'][] = array(
        'test_id' => $test['test_id'],
        'pass' => $test['status'] == PIFR_SERVER_TEST_RESULT_PASS, // TODO Fix once results codes are consistent.
        'message' => pifr_server_result_get_summary($test['test_id']),
      );
      pifr_server_log_event($test['test_id'], t('Result retrieved by project client.'));
      $count++;
    }

    return $batch;
  }
  else {
    pifr_debug('XML-RPC retrieve(): invalid test client.');
    return array(
      'response' => PIFR_RESPONSE_INVALID_SERVER,
    );
  }
  return array(
    'response' => PIFR_RESPONSE_ACCEPTED,
  );
}

/**
 * Convert a GMT Unix timestamp to a local Unix timestamp.
 *
 * @param integer $gmt_timestamp GMT Unix timestamp.
 * @return integer Local timestamp.
 */
function pifr_server_xmlrpc_time_local($gmt_timestamp) {
  $gmt_now = gmmktime();
  $offset = $gmt_now - $gmt_timestamp;
  return time() - $offset;
}

/**
 * Convert a local Unix timestamp to a GMT Unix timestamp.
 *
 * @param integer $timestamp Local Unix timestamp.
 * @return integer GMT Unix timestamp
 */
function pifr_server_xmlrpc_time_gmt($timestamp) {
  $now = time();
  $offset = $now - $timestamp;
  return gmmktime() - $offset;
}
